@@ -1,5 +1,5 @@
 """
-This module contains all core classes required for drawing diagrams.  
+This module contains all core classes required for drawing diagrams.
 
 Available Classes:
 - Graph
@@ -14,19 +14,20 @@ Available Functions:
 - set_graph
 - get_cluster
 - set_cluster
-- get_node
-- set_node
+- get_state
+- set_state
 - wrap_text
-- get_node_from_cluster
-- get_cluster_from_node
+- get_node_obj
+- get_cluster_obj
 
 Details for each can be found in the docstrings for the respective class or function.
 """
+from __future__ import annotations
 
 import contextvars
 import os
-import uuid
 from pathlib import Path
+from typing import Any, Union
 
 from graphviz import Digraph
 
@@ -34,56 +35,95 @@ from architectures.themes import Default
 
 __graph = contextvars.ContextVar("graph")
 __cluster = contextvars.ContextVar("cluster")
-__node = contextvars.ContextVar("node")
+__state = contextvars.ContextVar("state")
 
-def get_graph():
+
+def get_graph() -> Union[None, Graph]:
     """
-    Get the current graph context
+    Get the current graph context.
     """
     try:
         return __graph.get()
     except LookupError:
         return None
 
-def set_graph(graph):
+
+def set_graph(graph: Union[None, Graph]) -> None:
     """
-    Set the current graph context
+    Set the current graph context.
     """
     __graph.set(graph)
 
-def get_cluster():
+
+def get_cluster() -> Union[None, Cluster]:
     """
-    Get the current cluster context
+    Get the current cluster context.
     """
     try:
         return __cluster.get()
     except LookupError:
         return None
 
-def set_cluster(cluster):
+
+def set_cluster(cluster: Cluster) -> None:
     """
-    Set the current cluster context
+    Set the current cluster context.
     """
     __cluster.set(cluster)
 
-def get_node():
+
+def get_state() -> Union[None, dict]:
     """
-    Get the current node to cluster mapping
+    Get the current node to cluster mapping.
     """
     try:
-        return __node.get()
+        return __state.get()
     except LookupError:
         return None
 
-def set_node(node):
+
+def set_state(state: dict) -> None:
     """
-    Set a node to cluster mapping
+    Set a node to cluster mapping.
     """
-    __node.set(node)
+    __state.set(state)
 
 
-def wrap_text(text, max_length=16):
-    """Return a new label with wrapped text
+def update_state(state: dict, target_key: Union[Cluster, Node], target_value: Union[dict, Node]) -> dict:
+    """
+    Create a map of all resources hierarchial relationship as a nested dictionary.
+    """
+    if isinstance(state, dict):
+        for k, v in state.items():
+            if k is target_key:
+                v.append(target_value)
+            else:
+                if isinstance(v, list):
+                    for i in v:
+                        update_state(i, target_key, target_value)
+
+        return {k: v}
+
+
+def search_state(search_dict: dict, search_key: Cluster, output: list = []) -> list:
+    """
+    Search nested dicts and lists for the search_value (key)
+    and return the corresponding value.
+    """
+    
+    for k, v in search_dict.items():
+        if k is not search_key and isinstance(v, list):
+            for item in v:
+                if isinstance(item, dict):
+                    search_state(item, search_key, output)
+        else:
+            for item in v:
+                output.append(item)
+
+    return output
+
+def wrap_text(text: str, max_length: int = 16) -> str:
+    """Return a new label with wrapped text.
 
     Parameters
     ----------
@@ -99,7 +139,7 @@ def wrap_text(text, max_length=16):
     """
     if max_length < 12:
         max_length = 12
-        
+
     if len(text) > max_length:
         words = text.split()
         new_text = ""
@@ -117,48 +157,71 @@ def wrap_text(text, max_length=16):
     else:
         return text
 
-def get_node_from_cluster(cluster):
-    """Return the most central Node in a Cluster
+
+def get_node_obj(obj: Union[Cluster, Node]) -> Node:
+    """Return the most central Node in a Cluster.
 
     Parameters
     ----------
-    cluster : Cluster
-        A Cluster object
+    obj : Cluster, Node
+        A Cluster, Group, or Node object
 
     Returns
     -------
     Node
         The most centrally located Node object
     """
-    node_dict = get_node()
-    center_node_index = round(len(node_dict[cluster])/2) - 1
-    node = node_dict[cluster][center_node_index]
-    return node
+    if isinstance(obj, Cluster):
+        state = get_state()
+        values = search_state(state, obj, [])
+        print(values)
+        if any(isinstance(x, Node) for x in values):
+            node_list = [item for item in values if isinstance(item, Node)]
+            count = len(node_list)
+            center_node_index = round(count/2) - 1
+            obj = node_list[center_node_index]
+        else:
+            for item in values:
+                if isinstance(item, dict):
+                    for k, v in item.items():
+                        return get_node_obj(k)
+    
+    return obj
 
-def get_cluster_from_node(node):
-    """Return a Node's parent Cluster
+
+def get_cluster_obj(obj: Union[Cluster, Node]) -> Cluster:
+    """Return a Node's parent Cluster.
 
     Parameters
     ----------
-    node : Node
-        A Node object
+    obj : Cluster, Node
+        A Cluster, Group, or Node object
 
     Returns
     -------
     Cluster
         The parent Cluster
     """
-    node_dict = get_node()
-    for cluster, nodes in node_dict.items():
-        for item in nodes:
-            if item == node:
-                return cluster
+    if isinstance(obj, Node):
+        state = get_state()
+        for cluster, nodes in state.items():
+            for item in nodes:
+                if item == obj:
+                    return cluster
+    elif isinstance(obj, Cluster):
+        return obj
+    else:
+        raise TypeError("The Edge object only accepts Clusters, Groups, and Nodes.")
+
 
 class Graph():
     """
     Create and set default settings for a graph and its clusters, nodes, and edges.
     """
-    def __init__(self, name="my-architecture", output_file_format="png", theme=None, show=True):
+    def __init__(self, name: str = "my-architecture",
+                 output_file_format: str = "png",
+                 theme: Any = None, show: bool = True
+                 ) -> None:
         """
         :param str name: The name of the graph.
         :param str output_file_format: The format of the output file.
@@ -193,43 +256,47 @@ class Graph():
         # Set option to show architecture diagram
         self.show = show
 
-    def __str__(self):
-        return str(self.dot)
+        # Set initial state to just the Graph
+        set_state({self: []})
 
-    def __enter__(self):
+    # def __str__(self) -> str:
+    #     return str(self.dot)
+
+    def __enter__(self) -> Graph:
         set_graph(self)
         return self
 
-    def __exit__(self, exc_type, exc_value, traceback):
+    def __exit__(self, exc_type, exc_value, traceback) -> None:
         self.render()
         # Remove the graphviz file leaving only the image.
         os.remove(self.output_file_name)
         set_graph(None)
 
-    def node(self, node_id, label, **attrs):
+    def node(self, node_id: str, label: str, **attrs: Any) -> None:
         """
         Create a node.
         """
         self.dot.node(node_id, label=label, **attrs)
 
-    def edge(self, start_node, end_node, **attrs):
+    def edge(self, start_node: Node, end_node: Node, **attrs: Any) -> None:
         """
         Connect individual or lists of nodes with edges.
         """
         self.dot.edge(start_node.node_id, end_node.node_id, **attrs)
 
-    def subgraph(self, dot):
+    def subgraph(self, dot: Digraph) -> None:
         """
         Create a subgraph for grouping nodes.
         """
         self.dot.subgraph(dot)
 
-    def render(self):
+    def render(self) -> None:
         """
         Generate output file.
         """
         self.dot.render(format=self.output_file_format, view=self.show, quiet=True)
 
+
 class Cluster():
     """
     Create a cluster.
@@ -237,13 +304,13 @@ class Cluster():
 
     _default_label = None
 
-    def __init__(self, label="", **attrs):
+    def __init__(self, label: str = "", **attrs: Any) -> None:
         """
         :param label: Label for the cluster.
         """
 
         # Set the cluster name
-        self.name = "cluster_" + self._rand_id()
+        self.name = "cluster_" + str(id(self))
 
         #Set the cluster label
         if label == "" and self._default_label:
@@ -280,44 +347,47 @@ class Cluster():
         # Override any values directly passed from the object
         self.dot.graph_attr.update(attrs)
 
+        # Add Clusters to state
+        state = get_state()
+        if self._cluster:
+            state = update_state(state, self._cluster, {self: []})
+        else:
+            state = update_state(state, self._graph, {self: []})
+        set_state(state)
 
-    def __enter__(self):
+    def __enter__(self) -> Cluster:
         set_cluster(self)
         return self
 
-    def __exit__(self, exc_type, exc_value, traceback):
+    def __exit__(self, exc_type, exc_value, traceback) -> None:
         if self._cluster:
             self._cluster.subgraph(self.dot)
         else:
             self._graph.subgraph(self.dot)
         set_cluster(self._cluster)
 
-    def node(self, node_id, label, **attrs):
+    def node(self, node_id: str, label: str, **attrs: Any) -> None:
         """
         Create a node in the cluster.
         """
         self.dot.node(node_id, label=label, **attrs)
 
-    def subgraph(self, dot=Digraph):
+    def subgraph(self, dot: Digraph = Digraph) -> None:
         """
         Create a subgraph of the cluster.
         """
         self.dot.subgraph(dot)
 
-    @staticmethod
-    def _rand_id():
-        return uuid.uuid4().hex
-
 
 class Group(Cluster):
     """
     Creates a special type of cluster used only or grouping nodes.
     """
 
-    def __init__(self, label="group", **attrs):
+    def __init__(self, label: str = "group", **attrs: Any) -> None:
 
         # Set the group name
-        self.name = "cluster_" + self._rand_id()
+        self.name = "cluster_" + str(id(self))
 
         # Set the group label
         self.label = label
@@ -337,6 +407,15 @@ class Group(Cluster):
         # Set group depth
         self._depth = self._cluster._depth + 1 if self._cluster else 0
 
+        # Add Groups to state
+        state = get_state()
+        if self._cluster:
+            state = update_state(state, self._cluster, {self: []})
+        else:
+            state = update_state(state, self._graph, {self: []})
+        set_state(state)
+
+
 class Node():
     """
     Creates a node.  This can be a standard node or a node representing a service from a provider.
@@ -350,12 +429,15 @@ class Node():
 
     _default_label = None
 
-    def __init__(self, label="", wrap_label_text=True, **attrs):
+    def __init__(self, label: str = "",
+                 wrap_label_text: bool = True,
+                 **attrs: Any
+                 ) -> None:
         """
         :param str label: Label for a node.
         """
         # Generate an ID used to uniquely identify a node
-        self._id = self._rand_id()
+        self.id = str(id(self))
 
         #Set the label
         if self._icon and label == "":
@@ -398,133 +480,105 @@ class Node():
 
         # If a node is in the cluster context, add it to cluster.
         if self._cluster:
-            self._cluster.node(self._id, self.label, **self.node_attrs)
+            self._cluster.node(self.id, self.label, **self.node_attrs)
         else:
-            self._graph.node(self._id, self.label, **self.node_attrs)
+            self._graph.node(self.id, self.label, **self.node_attrs)
 
-        node_dict = get_node()
-        if node_dict is None:
-            # Creates the initial node dictionary if one does not exist
-            node_dict = {self._cluster: [self]}
-            set_node(node_dict)
-        elif self._cluster not in node_dict:
-            # Adds a new cluster key and value to the node dictionary
-            node_dict.update({self._cluster: [self]})
+        # Add Nodes to state
+        state = get_state()
+        if self._cluster:
+            state = update_state(state, self._cluster, self)
         else:
-            # Updates an existing cluster value in the node dictionary
-            node_list = node_dict[self._cluster]
-            node_list.append(self)
-            node_dict.update({self._cluster: node_list})
-            set_node(node_dict)
+            state = update_state(state, self._graph, self)
+        set_state(state)
 
     @property
-    def node_id(self):
+    def node_id(self) -> str:
         """
         Return the node id
         """
-        return self._id
+        return self.id
 
-    @staticmethod
-    def _rand_id():
-        return uuid.uuid4().hex
-
-    def _load_icon(self):
+    def _load_icon(self) -> str:
         basedir = Path(os.path.abspath(os.path.dirname(__file__)))
         return os.path.join(basedir.parent.parent, self._icon_dir, self._icon)
 
+
 class Edge():
     """
-    Creates an edge between two nodes
+    Creates an edge between two nodes.
     """
 
-    def __init__(self, start_node, end_node, **attrs):
+    def __init__(self, start_obj: Union[Cluster, Group, Node],
+                 end_obj: Union[Cluster, Group, Node],
+                 **attrs: Any
+                 ) -> None:
         """
         :param start: The origin cluster, group, or node object.
         :param end: The destination cluster, group, or node object.
         :param attrs: Other edge attributes.
         """
-
-        self.start_node = start_node
-        self.end_node = end_node
+        self.start_obj = start_obj
+        self.end_obj = end_obj
 
         # Get global graph and cluster context to ensure the node is part of the graph or cluster
         self._graph = get_graph()
         if self._graph is None:
             raise EnvironmentError("The object is not part of a Graph")
-        self._node = get_node()
+        self._state = get_state()
 
         # Set edge attributes based on the theme using copy to ensure the objects are independent
         self.edge_attrs = self._graph.theme.edge_attrs.copy()
 
-        if not isinstance(self.start_node, list):
-            self.start_node = [self.start_node]
+        if not isinstance(self.start_obj, list):
+            self.start_obj = [self.start_obj]
         
-        if not isinstance(self.end_node, list):
-            self.end_node = [self.end_node]
+        if not isinstance(self.end_obj, list):
+            self.end_obj = [self.end_obj]
 
-        # Ensure object passed is of the correct type
-        for nodes in [self.start_node, self.end_node]:
-            if not all(isinstance(node, (Cluster, Group, Node)) for node in nodes):
-                raise TypeError("The Edge object only accepts Clusters, Groups, and Nodes.")
-
-        start_node_list = self.start_node
-        end_node_list = self.end_node
+        start_obj_list = self.start_obj
+        end_obj_list = self.end_obj
 
         # Handle all cases
-        for current_start_node in start_node_list:
-            for current_end_node in end_node_list:
-                if isinstance(current_start_node, Node) and isinstance(current_end_node, Node):
-                    self.start_node = current_start_node
-                    self.end_node = current_end_node
-                    self.edge_attrs.update({"ltail": "", "lhead": ""})
-                elif isinstance(current_start_node, Node) and isinstance(current_end_node, (Cluster, Group)):
-                    start_cluster = get_cluster_from_node(current_start_node)
-                    end_cluster = current_end_node
-                    if start_cluster == end_cluster:
-                        self.start_node = None
-                        self.end_node = None
-                    else:
-                        self.start_node = current_start_node
-                        self.end_node = get_node_from_cluster(current_end_node)
-                        self.edge_attrs.update({"lhead": end_cluster.name})
-                elif isinstance(current_start_node, (Cluster, Group)) and isinstance(current_end_node, Node):
-                    start_cluster = current_start_node
-                    end_cluster = get_cluster_from_node(current_end_node)
-                    if start_cluster == end_cluster:
-                        self.start_node = None
-                        self.end_node = None
-                    else:
-                        self.start_node = get_node_from_cluster(current_start_node)
-                        self.end_node = current_end_node
-                        self.edge_attrs.update({"ltail": start_cluster.name})
-                elif isinstance(current_start_node, (Cluster, Group)) and isinstance(current_end_node, (Cluster, Group)):
-                    start_cluster = current_start_node
-                    end_cluster = current_end_node
-                    if start_cluster == end_cluster:
-                        self.start_node = None
-                        self.end_node = None
-                    else:
-                        self.start_node = get_node_from_cluster(current_start_node)
-                        self.end_node = get_node_from_cluster(current_end_node)
-                        self.edge_attrs.update({"ltail": start_cluster.name, "lhead": end_cluster.name})
+        for current_start_obj in start_obj_list:
+            for current_end_obj in end_obj_list:
+                self.start_cluster = get_cluster_obj(current_start_obj)
+                self.end_cluster = get_cluster_obj(current_end_obj)
+                self.start_node = get_node_obj(current_start_obj)
+                self.end_node = get_node_obj(current_end_obj)
+
+                # Cluster to Cluster connections
+                if isinstance(current_start_obj, Cluster) and isinstance(current_end_obj, Cluster):
+                    self.edge_attrs.update({"ltail": self.start_cluster.name, "lhead": self.end_cluster.name})
+                    self_reference = self.start_cluster == self.end_cluster
+                # Cluster to Node connections
+                elif isinstance(current_start_obj, Cluster) and isinstance(current_end_obj, Node):
+                    self.edge_attrs.update({"ltail": self.start_cluster.name})
+                    self_reference = self.start_cluster == self.end_cluster
+                # Node to Cluster connections
+                elif isinstance(current_start_obj, Node) and isinstance(current_end_obj, Cluster):
+                    self.edge_attrs.update({"lhead": self.end_cluster.name})
+                    self_reference = self.start_cluster == self.end_cluster
+                # Node to Node connections
                 else:
-                    assert isinstance(self.start_node, (Cluster, Group, Node))
-                    assert isinstance(self.end_node, (Cluster, Group, Node))
+                    self.edge_attrs.update({"ltail": "", "lhead": ""})
+                    self_reference = self.start_node == self.end_node
 
                 # Override any attributes directly passed from the object
                 self.edge_attrs.update(attrs)
 
-                if self.start_node is not None and self.end_node is not None:
+                if not self_reference:
                     self._graph.edge(self.start_node, self.end_node, **self.edge_attrs)
 
+
 class Flow():
     """
-    Another method of connecting nodes by allowing users to define a flow as a list
+    Another method of connecting nodes by allowing users to define a flow as a list.
     """
-    def __init__(self, nodes, **attrs):
+    def __init__(self, objs: list[Union[Cluster, Node]], **attrs: Any) -> None:
 
-        self.nodes = nodes
-        self.node_count = len(self.nodes)
+        self.objs = objs
+        self.obj_count = len(self.objs)
 
         # Get global graph and cluster context to ensure the node is part of the graph or cluster
         self._graph = get_graph()
@@ -538,19 +592,20 @@ class Flow():
         self.edge_attrs.update(attrs)
 
         # Make sure there is more than one node
-        if self.node_count > 1:
+        if self.obj_count > 1:
 
             # For every node in the list
-            for i in range(self.node_count):
+            for i in range(self.obj_count):
 
                 # Set the start and end node
-                if i < self.node_count - 1:
-                    self.start_node = self.nodes[i]
-                    self.end_node = self.nodes[i + 1]
+                if i < self.obj_count - 1:
+                    self.start_obj = self.objs[i]
+                    self.end_obj = self.objs[i + 1]
 
-                    Edge(self.start_node, self.end_node, **self.edge_attrs)
+                    Edge(self.start_obj, self.end_obj, **self.edge_attrs)
         else:
-            raise Exception('More than one node must be passed in the list to use the Flow object')
+            raise Exception('More than one object must be passed in the list to use Flow')
+
 
 Connection = Edge
 Container = Cluster
